
//		ConSearch.cpp :  TCC plugin to search for text in the console scrollback buffer


#define UNICODE						1
#define _UNICODE					1

#include <stdio.h>
#include <tchar.h>

#include <windows.h>
#include <psapi.h>
#include <dwmapi.h>

#include "PlugIn.h"
#include "TakeCmd.h"

#include "resource.h"


#define VER_MAJOR					0
#define VER_MINOR					35
#define VER_BUILD					0
#define VER_PATCH					3

#define PLUGIN_NAME					L"ConSearch"

#ifdef _WIN64
#define PLATFORM					L"x64"
#else
#define PLATFORM					L"x86"
#endif


#define MAX_ARG_LEN					8192
#define MAX_ARGS					1024
#define SHORT_BUF_LEN				1024
#define MAX_FILENAME_LEN			16384


#define DEFAULT_HOTKEY				0x10021

#define SEARCH_CASE_INSENSITIVE		0
#define SEARCH_CASE_SENSITIVE		1
#define SEARCH_REGEXP				2

#define DEFAULT_NOT_FOUND			0x0000
#define NOT_FOUND_HAND				0x0000
#define NOT_FOUND_ASTERISK			0x0001
#define NOT_FOUND_DEFAULT_BEEP		0x0002
#define NOT_FOUND_EXCLAMATION		0x0003
#define NOT_FOUND_SOUNDS			0x0003
#define NOT_FOUND_MESSAGEBOX		0x0080
#define NOT_FOUND_CLOSE				0x0100

#define DEFAULT_COLOR1				0x004f
#define DEFAULT_COLOR2				0x003f

#define DEFAULT_ACTIVE_ALPHA		0x00ff
#define DEFAULT_INACTIVE_ALPHA		0x00c0
#define ALPHA_MIN					0x0030
#define ALPHA_MAX					0x00ff
#define ALPHA_OPAQUE				0x00ff

#define DEFAULT_TIMER_ID			0x1138
#define REFOCUS_TIMER_ID			0x1139

#define SLOP_MAX					20
#define SLOP_DEFAULT				4

#define LF							0x0a

#define WM_UPDATE					0x8010

#define MAX_CONSOLE2_WINDOWS		32

#define ERRHANDLE					GetStdHandle( STD_ERROR_HANDLE )



wchar_t ConSearchHelpText[] = L"Open the search dialog and search the console buffer for text.\n\
\n\
CONSEARCH /A: /I /R /S /W: text  \r\n\
\n\
\t/A:\twrap all lines:  Y or N\n\
\t/I\tcase-insensitive search\n\
\t/R\tregular expression search\n\
\t/S\tcase-sensitive search\n\
\t/W:\twhole words only:  Y or N\n\
\ttext\ttext or regexp to search for  \n\
\n\
If you supply the search text, the dialog will open and automatically  \n\
search for the first match.  If you don\u2019t supply any text, the dialog\n\
will open but the search will not begin.\n";


wchar_t ConSearchOptsHelpText[] = L"Display and change plugin settings.\n\
\n\
CONSEARCHOPTS /D /E: /F: /H: /K: /L /R /S /T: /W:  \r\n\
\n\
\t/D\trestore original plugin defaults\n\
\t/E:\terror handling:  A, H, D, E, M, C\n\
\t/F:\tfade the search dialog when inactive\n\
\t/H:\thighlight colors (value or quoted string)\n\
\t/K:\tspecify a different hotkey\n\
\t/L\tlist all supported hotkeys\n\
\t/R\trestore saved defaults from the registry\n\
\t/S\tsave settings to the registry as the new defaults  \n\
\t/T:\tdefault search type at startup:  I, R, S\n\
\t/W:\talways try to put dialog near console window\n\
\t/X:\tload plugin in a Take Command tab\n\
\n\
You cannot combine /D and /R.  To disable the hotkey, use /K:NONE.  \n";


const wchar_t PluginRegKey[]				= L"Software\\JPPlugins\\ConSearch";
const wchar_t HotkeyRegValue[]				= L"Hotkey";
const wchar_t NotFoundRegValue[]			= L"Errors";
const wchar_t CleverRegValue[]				= L"AlwaysReposition";
const wchar_t Color1RegValue[]				= L"HighlightColor1";
const wchar_t Color2RegValue[]				= L"HighlightColor2";
const wchar_t DefaultSearchTypeRegValue[]	= L"DefaultSearchType";
const wchar_t DefaultWholeWordsRegValue[]	= L"DefaultWholeWords";
const wchar_t DefaultWrapAllRegValue[]		= L"DefaultWrapAllLines";
const wchar_t NoTakeCommandRegValue[]		= L"NoLoadUnderTakeCommand";
const wchar_t FadeOutRegValue[]				= L"FadeOut";
const wchar_t ActiveAlphaRegValue[]			= L"ActiveAlpha";
const wchar_t InactiveAlphaRegValue[]		= L"InactiveAlpha";



wchar_t ColorNames[8][8]					= { L"black", L"blue", L"green", L"cyan", L"red", L"magenta", L"yellow", L"white" };


const wchar_t PluginFeatures[]				= L"consearch,consearchopts,*key";

const wchar_t HelpWindowTitle[]				= L"ConSearch Plugin Help";

const wchar_t Console2Class[]				= L"Console_2_Main";
const wchar_t ConEmuClass[]					= L"VirtualConsoleClass";
const wchar_t TCmdClass[]					= L"Afx:00400000:8:";
const wchar_t TCmdClass2[]					= L"XTPMainFrame";
const int TCmdClassLen						= 15;



LPTSTR arg[MAX_ARGS];
int arglen[MAX_ARGS];
int numargs = 0;
wchar_t allargs[MAX_ARG_LEN] = L"";


#include "HotKeys.cpp"



//		Global variables:


HINSTANCE MyDLLHandle				= NULL;
DWORD MyThread						= NULL;
HHOOK HookHandle					= NULL;

HWND SearchDialogHandle				= NULL;

DWORD HotKey						= DEFAULT_HOTKEY;

wchar_t SearchText[MAX_ARG_LEN]		= L"";
int SearchType						= SEARCH_CASE_INSENSITIVE;
BOOL WholeWordOnly					= FALSE;
BOOL WrapAllLines					= FALSE;
size_t SearchCaret					= 0;
DWORD NotFoundMode					= DEFAULT_NOT_FOUND;
UINT_PTR TimerID					= 0;
unsigned int AutoSearchTimeout		= 400;

int DefaultSearchType				= SEARCH_CASE_INSENSITIVE;
BOOL DefaultWholeWords				= FALSE;
BOOL DefaultWrapAll					= FALSE;

DWORD HighlightColor1				= DEFAULT_COLOR1;
DWORD HighlightColor2				= DEFAULT_COLOR2;

unsigned int HighlightRow			= 0;
unsigned int HighlightColumn		= 0;
unsigned int HighlightLength		= 0;
wchar_t HLOldText[MAX_ARG_LEN]		= L"";
WORD HLOldColors[MAX_ARG_LEN];


int DialogX = 0, DialogY = 0;
BOOL FirstOpen = TRUE, AlwaysBeClever = FALSE;

HWND Console2WindowHandle			= NULL;

DWORD NoLoadUnderTCmd				= 0;
DWORD FadeOut						= TRUE;
DWORD ActiveAlpha					= DEFAULT_ACTIVE_ALPHA;
DWORD InactiveAlpha					= DEFAULT_INACTIVE_ALPHA;
BOOL DialogActive					= TRUE;



int DisplayErrorHeader()
{
	wchar_t shortbuf[MAX_ARG_LEN] = L"%_batch";

	Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"%s plugin: ", PLUGIN_NAME );

	ExpandVariables( shortbuf, 1 );
	if ( shortbuf[0] == '0' )
		return 0;

	wcscpy_s( shortbuf, MAX_ARG_LEN, L"%_batchname [%_batchline]  " );
	ExpandVariables( shortbuf, 1 );
	Qprintf( GetStdHandle(STD_ERROR_HANDLE), shortbuf );

	return 0;
}



int ForceForegroundWindow( HWND WindowHandle )
{
	//		http://www.tek-tips.com/faqs.cfm?fid=4262

	if ( !WindowHandle )
		return -1;

	DWORD ForegroundThread = GetWindowThreadProcessId( GetForegroundWindow(), NULL );
	DWORD AppThread = GetCurrentThreadId();

	if ( ForegroundThread != AppThread ) {
		AttachThreadInput( ForegroundThread, AppThread, TRUE );
		BringWindowToTop( WindowHandle );
		ShowWindow( WindowHandle, SW_RESTORE );
		SetForegroundWindow( WindowHandle );
		AttachThreadInput( ForegroundThread, AppThread, FALSE );
	}
	else {
		BringWindowToTop( WindowHandle );
		ShowWindow( WindowHandle, SW_RESTORE );
		SetForegroundWindow( WindowHandle );
	}

	return 0;
}



int ShowCmdHelp( LPCTSTR HelpText, BOOL ShowVersion )
{
	QPuts( (LPTSTR) HelpText );

	if ( ShowVersion )
		Printf( L"\r\n\t\t\t\t%s  %u.%u.%u.%u (%s)\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, VER_PATCH, PLATFORM );

	return 0;
}



int OffsetToRowCol( size_t Offset, unsigned int Rows, unsigned int *LineStarts )
{
	unsigned int Row = 0, Best = 0;

	for ( unsigned int i = 0; i < Rows; i++ ) {
		if ( ( LineStarts[i] >= Best ) && ( Offset >= LineStarts[i] ) ) {
			Row = i;
			Best = LineStarts[i];
		}
	}

	HighlightRow = Row;
	HighlightColumn = (unsigned int) Offset - LineStarts[Row];

	return 0;
}



int MoveViewport( HANDLE ConsoleBufferHandle, unsigned int Row, unsigned int Column )
{
	CONSOLE_SCREEN_BUFFER_INFO BufferInfo;
	SMALL_RECT Viewport;
	int ViewportWidth = 0, ViewportHeight = 0;

	GetConsoleScreenBufferInfo( ConsoleBufferHandle, &BufferInfo );
	Viewport = BufferInfo.srWindow;
	ViewportWidth = Viewport.Right - Viewport.Left;
	ViewportHeight = Viewport.Bottom - Viewport.Top;

	if ( Row < (unsigned int) Viewport.Top ) {
		Viewport.Top = Row ? ( Row - 1 ) : 0;
		Viewport.Bottom = Viewport.Top + ViewportHeight;
	}
	else if ( Row > (unsigned int) Viewport.Bottom ) {
		Viewport.Bottom = Row + 1;
		if ( Viewport.Bottom >= BufferInfo.dwSize.Y )
			Viewport.Bottom = BufferInfo.dwSize.Y - 1;

		Viewport.Top = Viewport.Bottom - ViewportHeight;
	}

	if ( Column < (unsigned int) Viewport.Left ) {
		Viewport.Left = Column - 10;
		if ( Viewport.Left < 0 )
			Viewport.Left = 0;

		Viewport.Right = Viewport.Left + ViewportWidth;
	}
	else if ( Column > (unsigned int) Viewport.Right ) {
		Viewport.Right = Column + ( ViewportWidth / 2 );
		if ( Viewport.Right >= BufferInfo.dwSize.X )
			Viewport.Right = BufferInfo.dwSize.X - 1;

		Viewport.Left = Viewport.Right - ViewportWidth;
	}

	SetConsoleWindowInfo( ConsoleBufferHandle, TRUE, &Viewport );
	return 0;
}



int UnhighlightText( HANDLE ConsoleBufferHandle )
{
	if ( !HighlightLength )
		return 0;

	BOOL CloseMe = FALSE;
	CONSOLE_SCREEN_BUFFER_INFO BufferInfo;
	
	if ( !ConsoleBufferHandle ) {
		ConsoleBufferHandle = CreateFile( L"CONOUT$", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
		GetConsoleScreenBufferInfo( ConsoleBufferHandle, &BufferInfo );
		MoveViewport( ConsoleBufferHandle, BufferInfo.dwCursorPosition.Y, BufferInfo.dwCursorPosition.X );
		CloseMe = TRUE;
	}

	COORD HLCoord = { HighlightColumn, HighlightRow };
	DWORD Dummy = 0;
	wchar_t TestText[MAX_ARG_LEN] = L"";

	ReadConsoleOutputCharacter( ConsoleBufferHandle, TestText, HighlightLength, HLCoord, &Dummy );
	TestText[HighlightLength] = '\0';

	if ( !wcscmp( TestText, HLOldText ) )
		WriteConsoleOutputAttribute( ConsoleBufferHandle, HLOldColors, HighlightLength, HLCoord, &Dummy );

	if ( CloseMe )
		CloseHandle( ConsoleBufferHandle );

	HighlightLength = 0;
	return 0;
}



int HighlightText( size_t Length, HANDLE ConsoleBufferHandle )
{
	if ( !Length )
		return 0;

	COORD HLCoord = { HighlightColumn, HighlightRow };
	DWORD Dummy = 0;
	WORD Temp[MAX_ARG_LEN];
	WORD HighlightColor = 0x0000;


	ReadConsoleOutputAttribute( ConsoleBufferHandle, HLOldColors, (DWORD) Length, HLCoord, &Dummy );
	ReadConsoleOutputCharacter( ConsoleBufferHandle, HLOldText, (DWORD) Length, HLCoord, &Dummy );
	HLOldText[Length] = '\0';


	if ( ( ( HighlightColor1 & 0x00f0 ) >> 4 ) == ( HighlightColor1 & 0x000f ) )
		HighlightColor = ( ( HLOldColors[0] & 0x00f0 ) >> 4 ) | ( ( HLOldColors[0] & 0x000f ) << 4 );
	else if ( ( HLOldColors[0] & 0x00f0 ) == ( HighlightColor1 & 0x00f0 ) )
		HighlightColor = HighlightColor2 & 0xff;
	else
		HighlightColor = HighlightColor1 & 0xff;


	for ( size_t i = 0; i < Length; i++ )
		Temp[i] = HighlightColor;

	WriteConsoleOutputAttribute( ConsoleBufferHandle, Temp, (DWORD) Length, HLCoord, &Dummy );
	HighlightLength = (unsigned int) Length;

	MoveViewport( ConsoleBufferHandle, HighlightRow, HighlightColumn );
	return 0;
}



unsigned int ConsoleBufferToString( HANDLE ConsoleBufferHandle, unsigned int Rows, unsigned int Columns, LPTSTR StringBuffer, unsigned int *LineStarts )
{
	if ( !ConsoleBufferHandle || !StringBuffer )
		return 0;

	unsigned int i = 0;

	wchar_t *TempBuffer = NULL;
	COORD ReadCoord = { 0, 0 };
	DWORD CharsRead = 0;
	int LastNonBlank = -1;
	unsigned int TruncateChar = 0;


	TempBuffer = (wchar_t*) malloc( ( Columns + 1 ) * sizeof( wchar_t ) );
	if ( !TempBuffer )
		return 0;


	StringBuffer[i] = '\0';
	TempBuffer[Columns] = '\0';

	for ( unsigned int Row = 0; Row < Rows; Row++ ) {
		LineStarts[Row] = i;

		ReadCoord.X = 0;
		ReadCoord.Y = Row;
		ReadConsoleOutputCharacter( ConsoleBufferHandle, TempBuffer, Columns, ReadCoord, &CharsRead );

		if ( WrapAllLines ) {
			wcscpy_s( StringBuffer + i, Columns + 1, TempBuffer );
			i += Columns;
		}
		else {
			LastNonBlank = -1;
			for ( int j = 0; j < (int) Columns; j++ )
				if ( !iswspace( TempBuffer[j] ) )
					LastNonBlank = j;

			if ( LastNonBlank == -1 ) {
				StringBuffer[i++] = LF;
				StringBuffer[i] = '\0';
			}
			else if ( LastNonBlank == ( Columns - 1 ) ) {
				wcscpy_s( StringBuffer + i, Columns + 1, TempBuffer );
				i += Columns;
			}
			else {
				wcsncpy_s( StringBuffer + i, Columns + 1, TempBuffer, LastNonBlank + 1 );
				i += ( LastNonBlank + 1 );
				StringBuffer[i++] = LF;
				StringBuffer[i] = '\0';
				TruncateChar = i;
			}
		}
	}

	free( TempBuffer );

	if ( TruncateChar ) {
		StringBuffer[TruncateChar] = '\0';
		return TruncateChar;
	}

	StringBuffer[i++] = LF;
	StringBuffer[i] = '\0';
	return i;
}



BOOL GoodMatch( LPTSTR StringBuffer, LPTSTR SearchString, size_t Offset )
{
	if ( !StringBuffer || !SearchString )
		return FALSE;

	if ( WholeWordOnly && Offset )
		if ( iswalnum( StringBuffer[Offset - 1] ) )
			return FALSE;

	size_t i = 0;

	while ( SearchString[i] && ( StringBuffer[Offset + i] == SearchString[i] ) )
		i++;

	if ( SearchString[i] )
		return FALSE;

	if ( WholeWordOnly )
		if ( iswalnum( StringBuffer[Offset + i] ) )
			return FALSE;

	return TRUE;
}



int ComplainNotFound( LPTSTR SearchString, BOOL Regexp )
{
	if ( NotFoundMode & NOT_FOUND_MESSAGEBOX ) {
		FmtMsgBox( FALSE, L"ConSearch Plugin", L"Search %s not found:\n%s", Regexp ? L"expression" : L"text", SearchString );
		
		if ( SearchDialogHandle )
			ForceForegroundWindow( SearchDialogHandle );
	}
	else if ( NotFoundMode & NOT_FOUND_CLOSE ) {
		if ( SearchDialogHandle )
			DestroyWindow( SearchDialogHandle );
	}
	else {
		switch ( NotFoundMode & NOT_FOUND_SOUNDS ) {
			case NOT_FOUND_ASTERISK :
				MessageBeep( MB_ICONASTERISK );
				break;

			case NOT_FOUND_DEFAULT_BEEP :
				MessageBeep( MB_OK );
				break;

			case NOT_FOUND_EXCLAMATION :
				MessageBeep( MB_ICONEXCLAMATION );
				break;

			default :
				MessageBeep( MB_ICONHAND );
				break;
		}
	}

	return 0;
}



int SearchForString( LPTSTR StringBuffer, LPTSTR SearchText, BOOL CaseSensitive, unsigned int BufferChars, unsigned int *LineStarts, BOOL Reverse, unsigned int Rows, HANDLE ConsoleBufferHandle )
{
	if ( !StringBuffer || !SearchText )
		return -1;

	if ( ( StringBuffer[0] == '\0' ) || ( SearchText == '\0' ) )
		return -1;


	wchar_t SearchString[MAX_ARG_LEN] = L"";
	size_t Offset = 0;
	size_t len = wcslen( SearchText );


	wcscpy_s( SearchString, MAX_ARG_LEN, SearchText );

	if ( !CaseSensitive ) {
		_wcslwr_s( SearchString, MAX_ARG_LEN );
		_wcslwr_s( StringBuffer, BufferChars + 1 );
	}


	if ( Reverse ) {
		if ( SearchCaret > len ) {							//	Search backward from the caret to the start of the buffer
			for ( Offset = SearchCaret - len - 1; Offset != -1; Offset-- ) {
				if ( StringBuffer[Offset] == SearchString[0] )
					if ( GoodMatch( StringBuffer, SearchString, Offset ) ) {
						SearchCaret = Offset + len;
						if ( SearchCaret >= BufferChars )
							SearchCaret = 0;

						OffsetToRowCol( Offset, Rows, LineStarts );
						HighlightText( len, ConsoleBufferHandle );
						return 0;
					}
			}
		}

		if ( SearchCaret < ( BufferChars - len ) ) {		//	Search backward from the end of the buffer to the caret
			size_t min = 0;
			if ( SearchCaret >= len )
				min = SearchCaret - len;

			for ( Offset = BufferChars - len - 1; Offset != ( min - 1 ); Offset-- ) {
				if ( StringBuffer[Offset] == SearchString[0] )
					if ( GoodMatch( StringBuffer, SearchString, Offset ) ) {
						SearchCaret = Offset + len;
						if ( SearchCaret >= BufferChars )
							SearchCaret = 0;

						OffsetToRowCol( Offset, Rows, LineStarts );
						HighlightText( len, ConsoleBufferHandle );
						return 0;
					}
			}
		}
	}
	else {
		if ( SearchCaret < BufferChars ) {					//	Search forward from the caret to the end of the buffer
			for ( Offset = SearchCaret; Offset < BufferChars; Offset++ ) {
				if ( StringBuffer[Offset] == SearchString[0] )
					if ( GoodMatch( StringBuffer, SearchString, Offset ) ) {
						SearchCaret = Offset + len;
						if ( SearchCaret >= BufferChars )
							SearchCaret = 0;

						OffsetToRowCol( Offset, Rows, LineStarts );
						HighlightText( len, ConsoleBufferHandle );
						return 0;
					}
			}
		}

		if ( SearchCaret ) {								//	Search foward from the start of the buffer to the caret
			for ( Offset = 0; Offset <= SearchCaret; Offset++ ) {
				if ( StringBuffer[Offset] == SearchString[0] )
					if ( GoodMatch( StringBuffer, SearchString, Offset ) ) {
						SearchCaret = Offset + len;
						if ( SearchCaret >= BufferChars )
							SearchCaret = 0;

						OffsetToRowCol( Offset, Rows, LineStarts );
						HighlightText( len, ConsoleBufferHandle );
						return 0;
					}
			}
		}
	}

	return 1;
}



int SearchForRegexp( LPTSTR StringBuffer, LPTSTR Regexp, unsigned int BufferChars, unsigned int *LineStarts, BOOL Reverse, unsigned int Rows, HANDLE ConsoleBufferHandle )
{
	if ( !StringBuffer || !Regexp )
		return -1;
	else if ( ( StringBuffer[0] == '\0' ) || ( Regexp[0] == '\0' ) )
		return -1;


	LPTSTR MatchStart		= NULL;
	LPTSTR MatchEnd			= NULL;
	unsigned int Region		= 0;


	if ( Reverse ) {
		if ( SearchCaret ) {					//	Search backward from the caret to the start of the buffer
			MatchStart	= StringBuffer + SearchCaret - 1;
			MatchEnd	= StringBuffer;

			if ( !RegularExpression( Regexp, (BYTE*) StringBuffer, (BYTE**) &MatchStart, (BYTE**) &MatchEnd, &Region, 0, TRUE, TRUE ) )
				if ( MatchStart && MatchEnd ) {
					SearchCaret = ( MatchEnd - StringBuffer );
					if ( SearchCaret >= BufferChars )
						SearchCaret = 0;

					OffsetToRowCol( MatchStart - StringBuffer, Rows, LineStarts );
					HighlightText( MatchEnd - MatchStart, ConsoleBufferHandle );
					return 0;
				}
		}

		if ( SearchCaret < BufferChars ) {		//	Search backward from the end of the buffer to the caret
			MatchStart	= StringBuffer + BufferChars;
			MatchEnd	= StringBuffer + SearchCaret;

			if ( !RegularExpression( Regexp, (BYTE*) StringBuffer, (BYTE**) &MatchStart, (BYTE**) &MatchEnd, &Region, 0, TRUE, TRUE ) )
				if ( MatchStart && MatchEnd ) {
					SearchCaret = ( MatchEnd - StringBuffer );
					if ( SearchCaret >= BufferChars )
						SearchCaret = 0;

					OffsetToRowCol( MatchStart - StringBuffer, Rows, LineStarts );
					HighlightText( MatchEnd - MatchStart, ConsoleBufferHandle );
					return 0;
				}
		}
	}
	else {
		if ( SearchCaret < BufferChars ) {		//	Search forward from the caret to the end of the buffer
			MatchStart	= StringBuffer + SearchCaret;
			MatchEnd	= StringBuffer + BufferChars;

			if ( !RegularExpression( Regexp, (BYTE*) StringBuffer, (BYTE**) &MatchStart, (BYTE**) &MatchEnd, &Region, 0, TRUE, TRUE ) )
				if ( MatchStart && MatchEnd ) {
					SearchCaret = ( MatchStart - StringBuffer + 1 );
					if ( SearchCaret >= BufferChars )
						SearchCaret = 0;

					OffsetToRowCol( MatchStart - StringBuffer, Rows, LineStarts );
					HighlightText( MatchEnd - MatchStart, ConsoleBufferHandle );
					return 0;
				}
		}
		
		if ( SearchCaret ) {					//	Search forward from the start of the buffer
			MatchStart	= StringBuffer;
			MatchEnd	= StringBuffer + BufferChars;

			if ( !RegularExpression( Regexp, (BYTE*) StringBuffer, (BYTE**) &MatchStart, (BYTE**) &MatchEnd, &Region, 0, TRUE, TRUE ) )
				if ( MatchStart && MatchEnd ) {
					SearchCaret = ( MatchStart - StringBuffer + 1 );
					if ( SearchCaret >= BufferChars )
						SearchCaret = 0;

					OffsetToRowCol( MatchStart - StringBuffer, Rows, LineStarts );
					HighlightText( MatchEnd - MatchStart, ConsoleBufferHandle );
					return 0;
				}
		}
	}

	return 1;
}


	
int PerformSearch( LPTSTR SearchText, BOOL Reverse )
{
	if ( !SearchText )
		return -1;
	else if ( SearchText[0] == '\0' )
		return -1;


	HANDLE ConsoleBufferHandle = NULL;
	CONSOLE_SCREEN_BUFFER_INFO BufferInfo;
	wchar_t *StringBuffer = NULL;
	unsigned int StringBufferSize = 0;
	unsigned int *LineStarts = NULL;
	unsigned int BufferChars = 0;


	ConsoleBufferHandle = CreateFile( L"CONOUT$", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );

	if ( ConsoleBufferHandle && ( ConsoleBufferHandle != INVALID_HANDLE_VALUE ) ) {
		UnhighlightText( ConsoleBufferHandle );

		if ( GetConsoleScreenBufferInfo( ConsoleBufferHandle, &BufferInfo ) ) {

			StringBufferSize = BufferInfo.dwSize.Y * ( BufferInfo.dwSize.X + 1 ) + 2;
			StringBuffer = (wchar_t*) malloc( StringBufferSize * sizeof( wchar_t ) );

			LineStarts = (unsigned int*) malloc( ( BufferInfo.dwSize.Y + 1 ) * sizeof( unsigned int ) );

			if ( StringBuffer && LineStarts ) {
				memset( LineStarts, 0x00, ( BufferInfo.dwSize.Y + 1 ) * sizeof( unsigned int ) );

				BufferChars = ConsoleBufferToString( ConsoleBufferHandle, BufferInfo.dwSize.Y, BufferInfo.dwSize.X, StringBuffer, LineStarts );
				if ( SearchCaret >= BufferChars )
					SearchCaret = 0;

				if ( SearchType == SEARCH_REGEXP ) {
					if ( SearchForRegexp( StringBuffer, SearchText, BufferChars, LineStarts, Reverse, BufferInfo.dwSize.Y, ConsoleBufferHandle ) )
						ComplainNotFound( SearchText, TRUE );
				}
				else if ( SearchForString( StringBuffer, SearchText, ( SearchType == SEARCH_CASE_SENSITIVE ), BufferChars, LineStarts, Reverse, BufferInfo.dwSize.Y, ConsoleBufferHandle ) )
					ComplainNotFound( SearchText, FALSE );

			}
			free( StringBuffer );
			free( LineStarts );
		}

		CloseHandle( ConsoleBufferHandle );
	}

	return 0;
}



//	Returns the filename, ignoring any drive or path specification.

LPTSTR _FilenamePart( LPTSTR Filename, LPTSTR outFilename )
{
	if ( !Filename )
		return NULL;
	else if ( Filename[0] == '\0' )
		return NULL;


	BOOL FoundSlash			= FALSE;
	size_t EndOfPathPart	= -1;


	for ( size_t i = 0; Filename[i]; i++ ) {
		if ( ( Filename[i] == '/' ) || ( Filename[i] == '\\' ) ) {
			EndOfPathPart = i;
			FoundSlash = TRUE;
		}
		else if ( ( Filename[i] == ':' ) && !FoundSlash )
			EndOfPathPart = i;
	}

	if ( EndOfPathPart == -1 ) {
		if ( outFilename ) {
			wcscpy_s( outFilename, MAX_FILENAME_LEN, Filename );
			return outFilename;
		}
		else
			return Filename;
	}

	if ( outFilename ) {
		wcscpy_s( outFilename, MAX_FILENAME_LEN, Filename + EndOfPathPart + 1 );
		return outFilename;
	}
	else
		return Filename + EndOfPathPart + 1;
}



LRESULT CALLBACK MyMessageFilter( int Code, WPARAM wParam, LPARAM lParam )
{
	if ( Code < 0 )
		return CallNextHookEx( HookHandle, Code, wParam, lParam );

	MSG *Message = (MSG*) lParam;

	if ( SearchDialogHandle && Message->hwnd ) {
		if ( IsDialogMessage( SearchDialogHandle, Message ) ) {
			//  wprintf( L"   * IsDialogMessage() returned nonzero.\n" );

			if ( ( Message->message & 0xffffff00 ) == 0x0100 )
				Message->message = WM_NULL;

			return 0;
		}
	}

	return CallNextHookEx( HookHandle, Code, wParam, lParam );
}



INT_PTR CALLBACK SearchDialogProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	switch ( uMsg ) {

	case WM_INITDIALOG:

		SearchCaret = 0;
		DialogActive = TRUE;
		//  fall through ....

	case WM_UPDATE:

		switch ( SearchType ) {

		case SEARCH_CASE_INSENSITIVE :

			CheckRadioButton( hWnd, IDC_RADIO1, IDC_RADIO3, IDC_RADIO1 );
			EnableWindow( GetDlgItem( hWnd, IDC_CHECK1 ), TRUE );
			break;

		case SEARCH_CASE_SENSITIVE :

			CheckRadioButton( hWnd, IDC_RADIO1, IDC_RADIO3, IDC_RADIO2 );
			EnableWindow( GetDlgItem( hWnd, IDC_CHECK1 ), TRUE );
			break;

		case SEARCH_REGEXP :

			CheckRadioButton( hWnd, IDC_RADIO1, IDC_RADIO3, IDC_RADIO3 );
			EnableWindow( GetDlgItem( hWnd, IDC_CHECK1 ), FALSE );
			break;
		}

		if ( WholeWordOnly )
			CheckDlgButton( hWnd, IDC_CHECK1, BST_CHECKED );
		else
			CheckDlgButton( hWnd, IDC_CHECK1, BST_UNCHECKED );

		if ( WrapAllLines )
			CheckDlgButton( hWnd, IDC_CHECK2, BST_CHECKED );
		else
			CheckDlgButton( hWnd, IDC_CHECK2, BST_UNCHECKED );

		if ( SearchText[0] )
			SetDlgItemText( hWnd, IDC_EDIT1, SearchText );

		return TRUE;

	case WM_COMMAND :

		switch LOWORD( wParam ) {

		case IDCANCEL :

			DestroyWindow( hWnd );
			return TRUE;

		case IDOK :

			SearchText[0] = '\0';
			GetDlgItemText( hWnd, IDC_EDIT1, SearchText, MAX_ARG_LEN );

			if ( SearchText[0] )
				PerformSearch( SearchText, FALSE );

			return TRUE;

		case IDC_BUTTON1 :

			SearchText[0] = '\0';
			GetDlgItemText( hWnd, IDC_EDIT1, SearchText, MAX_ARG_LEN );

			if ( SearchText[0] )
				PerformSearch( SearchText, TRUE );

			return TRUE;

		case IDC_RADIO1 :

			SearchType = SEARCH_CASE_INSENSITIVE;
			EnableWindow( GetDlgItem( hWnd, IDC_CHECK1 ), TRUE );
			return TRUE;

		case IDC_RADIO2 :

			SearchType = SEARCH_CASE_SENSITIVE;
			EnableWindow( GetDlgItem( hWnd, IDC_CHECK1 ), TRUE );
			return TRUE;

		case IDC_RADIO3 :

			SearchType = SEARCH_REGEXP;
			EnableWindow( GetDlgItem( hWnd, IDC_CHECK1 ), FALSE );
			return TRUE;

		case IDC_CHECK1 :

			WholeWordOnly = ( IsDlgButtonChecked( hWnd, IDC_CHECK1 ) == BST_CHECKED );
			return TRUE;

		case IDC_CHECK2 :

			WrapAllLines = ( IsDlgButtonChecked( hWnd, IDC_CHECK2 ) == BST_CHECKED );
			return TRUE;
		}
		return TRUE;

	case WM_TIMER :

		KillTimer( hWnd, TimerID );
		PostMessage( hWnd, WM_COMMAND, IDOK, 0 );
		return FALSE;

	case WM_ACTIVATE :

		if ( LOWORD( wParam ) ) {
			//	wprintf( L"   * Activate:  DialogActive = %d\n", DialogActive );
			if ( FadeOut && !DialogActive )
				SetLayeredWindowAttributes( hWnd, 0, (BYTE) ActiveAlpha, LWA_ALPHA );

			DialogActive = TRUE;
		}
		else {
			//	wprintf( L"   * Deactivate:  DialogActive = %d\n", DialogActive );
			if ( FadeOut && DialogActive )
				SetLayeredWindowAttributes( hWnd, 0, (BYTE) InactiveAlpha, LWA_ALPHA );

			DialogActive = FALSE;
		}
		return FALSE;

	case WM_DESTROY :

		{
			UnhighlightText( NULL );

			RECT WindowFrame;
			GetWindowRect( hWnd, &WindowFrame );
			DialogX = WindowFrame.left;
			DialogY = WindowFrame.top;
				
			SearchText[0] = '\0';
			GetDlgItemText( hWnd, IDC_EDIT1, SearchText, MAX_ARG_LEN );

			PostQuitMessage( 0 );
			SearchDialogHandle = NULL;
		}
		return FALSE;

	default :

		return FALSE;
	}

	return FALSE;
}



int WindowOffScreen( HWND hWnd )
{
	int ScreenLeft = 0, ScreenTop = 0, ScreenWidth = 0, ScreenHeight = 0;
	int ScreenRight = 0, ScreenBottom = 0, SlopX = 0, SlopY = 0;

	ScreenLeft		= GetSystemMetrics( SM_XVIRTUALSCREEN );
	ScreenTop		= GetSystemMetrics( SM_YVIRTUALSCREEN );
	ScreenWidth		= GetSystemMetrics( SM_CXVIRTUALSCREEN );
	ScreenHeight	= GetSystemMetrics( SM_CYVIRTUALSCREEN );
	ScreenRight		= ScreenLeft + ScreenWidth;
	ScreenBottom	= ScreenTop + ScreenHeight;

	SlopX			= GetSystemMetrics( SM_CXSIZEFRAME );
	SlopY			= GetSystemMetrics( SM_CYSIZEFRAME );

	if ( SlopX > SLOP_MAX )
		SlopX = SLOP_DEFAULT;

	if ( SlopY > SLOP_MAX )
		SlopY = SLOP_DEFAULT;



	RECT WindowFrame = { 0, 0, 0, 0 };

	GetWindowRect( hWnd, &WindowFrame );

	if ( ( WindowFrame.left < ( ScreenLeft - SlopX ) ) || ( WindowFrame.top < ( ScreenTop - SlopY ) ) )
		return 1;

	if ( ( WindowFrame.right > ( ScreenRight + SlopX ) ) || ( WindowFrame.bottom > ( ScreenBottom + SlopY ) ) )
		return 1;

	return 0;
}



int CenterWindowOnMonitor( HWND hWnd )
{
	HMONITOR MonitorHandle = NULL;
	MONITORINFO MonitorInfo;
	RECT PopupFrame;
	int MonitorWidth = 0, MonitorHeight = 0, PopupWidth = 0, PopupHeight = 0;

	MonitorHandle = MonitorFromWindow( hWnd, MONITOR_DEFAULTTOPRIMARY );
	MonitorInfo.cbSize = sizeof( MONITORINFO );
	GetMonitorInfo( MonitorHandle, &MonitorInfo );

	GetWindowRect( hWnd, &PopupFrame );

	MonitorWidth	= MonitorInfo.rcWork.right	- MonitorInfo.rcWork.left;
	MonitorHeight	= MonitorInfo.rcWork.bottom	- MonitorInfo.rcWork.top;
	PopupWidth		= PopupFrame.right	- PopupFrame.left;
	PopupHeight		= PopupFrame.bottom	- PopupFrame.top;

	MoveWindow( hWnd, MonitorInfo.rcWork.left + ( ( MonitorWidth - PopupWidth ) / 2 ), MonitorInfo.rcWork.top + ( ( MonitorHeight - PopupHeight ) / 2 ), PopupWidth, PopupHeight, TRUE );
	return 0;
}



int CenterWindowOnWindow( HWND WindowToCenter, HWND MainWindow )
{
	if ( !MainWindow )
		return CenterWindowOnMonitor( WindowToCenter );
	else if ( WindowOffScreen( MainWindow ) )
		return CenterWindowOnMonitor( WindowToCenter );
	/*
	else if ( !IsWindowVisible( MainWindow ) )
		return CenterWindowOnMonitor( WindowToCenter );
	*/

	RECT MainFrame, PopupFrame;
	int MainWidth = 0, MainHeight = 0, PopupWidth = 0, PopupHeight = 0;

	GetWindowRect( MainWindow, &MainFrame );
	GetWindowRect( WindowToCenter, &PopupFrame );

	MainWidth	= MainFrame.right - MainFrame.left;
	MainHeight	= MainFrame.bottom - MainFrame.top;
	PopupWidth	= PopupFrame.right - PopupFrame.left;
	PopupHeight	= PopupFrame.bottom - PopupFrame.top;

	MoveWindow( WindowToCenter, MainFrame.left + ( ( MainWidth - PopupWidth ) / 2 ), MainFrame.top + ( ( MainHeight - PopupHeight ) / 2 ), PopupWidth, PopupHeight, TRUE );
	return 0;
}



int CleverPosition( HWND PopupWindow, HWND MainWindow )
{
	if ( !PopupWindow || !MainWindow )
		return -1;

	if ( !IsWindowVisible( MainWindow ) )
		return 1;

	RECT MainFrame, PopupFrame;
	int PopupWidth = 0, PopupHeight = 0;
	HMONITOR MonitorHandle = NULL;
	MONITORINFO MonitorInfo;
	int rc = 1;
	
	GetWindowRect( MainWindow, &MainFrame );
	GetWindowRect( PopupWindow, &PopupFrame );

	PopupWidth	= PopupFrame.right - PopupFrame.left;
	PopupHeight	= PopupFrame.bottom - PopupFrame.top;

	MonitorHandle = MonitorFromWindow( MainWindow, MONITOR_DEFAULTTONULL );
	if ( !MonitorHandle )
		return 1;

	MonitorInfo.cbSize = sizeof( MONITORINFO );
	GetMonitorInfo( MonitorHandle, &MonitorInfo );


	//  wprintf( L"   * MonitorInfo.rcWork = (%d, %d) - (%d, %d)\n", MonitorInfo.rcWork.left, MonitorInfo.rcWork.top, MonitorInfo.rcWork.right, MonitorInfo.rcWork.bottom );
	//  wprintf( L"   * MainFrame = (%d, %d) - (%d, %d)\n", MainFrame.left, MainFrame.top, MainFrame.right, MainFrame.bottom );

	if ( ( ( MainFrame.top - PopupHeight ) >= MonitorInfo.rcWork.top ) && ( MainFrame.top < MonitorInfo.rcWork.bottom ) ) {
		if ( ( MainFrame.left >= MonitorInfo.rcWork.left ) && ( ( MainFrame.left + PopupWidth ) < MonitorInfo.rcWork.right ) ) {
			MoveWindow( PopupWindow, MainFrame.left, MainFrame.top - PopupHeight, PopupWidth, PopupHeight, TRUE );
			rc = 0;
		}
		else if ( ( ( MainFrame.right - PopupWidth ) >= MonitorInfo.rcWork.left ) && ( MainFrame.right < MonitorInfo.rcWork.right ) ) {
			MoveWindow( PopupWindow, MainFrame.right - PopupWidth, MainFrame.top - PopupHeight, PopupWidth, PopupHeight, TRUE );
			rc = 0;
		}
	}
	else if ( ( MainFrame.bottom >= MonitorInfo.rcWork.top ) && ( ( MainFrame.bottom + PopupHeight ) < MonitorInfo.rcWork.bottom ) ) {
		if ( ( MainFrame.left >= MonitorInfo.rcWork.left ) && ( ( MainFrame.left + PopupWidth ) < MonitorInfo.rcWork.right ) ) {
			MoveWindow( PopupWindow, MainFrame.left, MainFrame.bottom, PopupWidth, PopupHeight, TRUE );
			rc = 0;
		}
		else if ( ( ( MainFrame.right - PopupWidth ) >= MonitorInfo.rcWork.left ) && ( MainFrame.right < MonitorInfo.rcWork.right ) ) {
			MoveWindow( PopupWindow, MainFrame.right - PopupWidth, MainFrame.bottom, PopupWidth, PopupHeight, TRUE );
			rc = 0;
		}
	}
	else if ( ( ( MainFrame.left - PopupWidth ) >= MonitorInfo.rcWork.left ) && ( MainFrame.left < MonitorInfo.rcWork.right ) ) {
		if ( ( MainFrame.top >= MonitorInfo.rcWork.top ) && ( ( MainFrame.top + PopupHeight ) < MonitorInfo.rcWork.bottom ) ) {
			MoveWindow( PopupWindow, MainFrame.left - PopupWidth, MainFrame.top, PopupWidth, PopupHeight, TRUE );
			rc = 0;
		}
		else if ( ( ( MainFrame.bottom - PopupHeight ) >= MonitorInfo.rcWork.top ) && ( MainFrame.bottom < MonitorInfo.rcWork.bottom ) ) {
			MoveWindow( PopupWindow, MainFrame.left - PopupWidth, MainFrame.bottom - PopupHeight, PopupWidth, PopupHeight, TRUE );
			rc = 0;
		}
	}
	else if ( ( MainFrame.right >= MonitorInfo.rcWork.left ) && ( ( MainFrame.right + PopupWidth ) < MonitorInfo.rcWork.right ) ) {
		if ( ( MainFrame.top >= MonitorInfo.rcWork.top ) && ( ( MainFrame.top + PopupHeight ) < MonitorInfo.rcWork.bottom ) ) {
			MoveWindow( PopupWindow, MainFrame.right, MainFrame.top, PopupWidth, PopupHeight, TRUE );
			rc = 0;
		}
		else if ( ( ( MainFrame.bottom - PopupHeight ) >= MonitorInfo.rcWork.top ) && ( MainFrame.bottom < MonitorInfo.rcWork.bottom ) ) {
			MoveWindow( PopupWindow, MainFrame.right, MainFrame.bottom - PopupHeight, PopupWidth, PopupHeight, TRUE );
			rc = 0;
		}
	}

	return rc;
}



int DummyConsoleOutput( unsigned int n )
{
	for ( unsigned int i = 0; i < n; i++ )
		_cputws( L"A\bB\bC\bD\b \b" );

	return 0;
}



BOOL CALLBACK TakeCmdWindowCheck( HWND hWnd, LPARAM lParam )
{
	if ( !lParam )
		return TRUE;


	wchar_t ClassName[SHORT_BUF_LEN] = L"";

	if ( !RealGetWindowClass( hWnd, ClassName, SHORT_BUF_LEN ) )
		return TRUE;

	if ( _wcsnicmp( ClassName, TCmdClass, TCmdClassLen ) )
		if ( _wcsicmp( ClassName, TCmdClass2 ) )
			return TRUE;

	HWND *TCmdHWnd = (HWND*) lParam;

	for ( int i = 0; i < MAX_CONSOLE2_WINDOWS; i++ ) {
		if ( TCmdHWnd[i] == NULL ) {
			TCmdHWnd[i] = hWnd;
			return TRUE;
		}
	}

	return TRUE;
}



HWND FindTakeCmdWindow()
{
	HWND TCmdHWnd[MAX_CONSOLE2_WINDOWS];
	DWORD TCmdPid[MAX_CONSOLE2_WINDOWS];

	for ( int i = 0; i < MAX_CONSOLE2_WINDOWS; i++ )
		TCmdHWnd[i] = NULL, TCmdPid[i] = 0;


	EnumWindows( TakeCmdWindowCheck, (LPARAM) TCmdHWnd );

	if ( TCmdHWnd[0] == NULL )
		return NULL;

	for ( int i = 0; i < MAX_CONSOLE2_WINDOWS; i++ )
		if ( TCmdHWnd[i] )
			GetWindowThreadProcessId( TCmdHWnd[i], TCmdPid + i );

	DWORD BufferSize			= 64;
	DWORD ProcessCount			= BufferSize + 1;
	DWORD *ProcessListBuffer	= NULL;


	DummyConsoleOutput( 256 );

	while ( ProcessCount > BufferSize ) {

		if ( ProcessListBuffer ) {
			free( ProcessListBuffer );
			ProcessListBuffer = NULL;
		}

		ProcessListBuffer = (DWORD*) malloc( BufferSize * sizeof( DWORD ) );

		if ( !ProcessListBuffer )
			return NULL;

		ProcessCount = GetConsoleProcessList( ProcessListBuffer, BufferSize );
	}

	if ( !ProcessCount ) {
		if ( ProcessListBuffer )
			free( ProcessListBuffer );

		return NULL;
	}

	for ( unsigned int i = 0; i < ProcessCount; i++ ) {
		if ( ProcessListBuffer[i] ) {
			for ( int j = 0; j < MAX_CONSOLE2_WINDOWS; j++ )
				if ( ProcessListBuffer[i] == TCmdPid[j] )
					return TCmdHWnd[j];
		}
	}

	return NULL;
}



int ParseInt( LPTSTR inString, unsigned int *outValue, wchar_t **EndChar )
{
	if ( !inString || !outValue )
		return -1;

	unsigned int value = 0;
	wchar_t *sep1 = NULL;

	_set_errno( 0 );

	if ( ( inString[0] == '0' ) && ( towupper( inString[1] ) == 'X' ) && ( iswxdigit( inString[2] ) ) ) {
		//  parse hexadecimal

		value = wcstoul( inString + 2, &sep1, 16 );

		if ( errno )
			return -1;
		else if ( !EndChar && *sep1 )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}
	else if ( iswdigit( inString[0] ) ) {
		//  parse unsigned

		value = wcstoul( inString, &sep1, 10 );

		if ( errno )
			return -1;
		else if ( !EndChar && *sep1 )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}

	return -1;
}



int GetConEmuCFilename( LPTSTR outString )
{
	if ( !outString )
		return -1;


	DWORD BufferSize			= 64;
	DWORD ProcessCount			= BufferSize + 1;
	DWORD *ProcessListBuffer	= NULL;


	while ( ProcessCount > BufferSize ) {

		if ( ProcessListBuffer ) {
			free( ProcessListBuffer );
			ProcessListBuffer = NULL;
		}

		ProcessListBuffer = (DWORD*) malloc( BufferSize * sizeof( DWORD ) );

		if ( !ProcessListBuffer )
			return 1;

		ProcessCount = GetConsoleProcessList( ProcessListBuffer, BufferSize );
	}

	if ( !ProcessCount ) {
		if ( ProcessListBuffer )
			free( ProcessListBuffer );

		return 1;
	}


	wchar_t Filename[MAX_FILENAME_LEN]			= L"";
	LPTSTR BaseName			= NULL;
	HANDLE ProcessHandle	= NULL;
	int rv					= 1;


	for ( unsigned int i = 0; i < ProcessCount; i++ ) {
		Filename[0] = '\0';
		
		ProcessHandle = OpenProcess( PROCESS_QUERY_INFORMATION | PROCESS_VM_READ, FALSE, ProcessListBuffer[i] );

		if ( ProcessHandle ) {
			GetModuleFileNameEx( ProcessHandle, NULL, Filename, MAX_FILENAME_LEN );
			CloseHandle( ProcessHandle );
			ProcessHandle = NULL;
		}

		if ( Filename[0] ) {
			BaseName = _FilenamePart( Filename, NULL );

			if ( BaseName ) {
				if ( !_wcsicmp( BaseName, L"ConEmuC64.exe" ) ) {
					wcscpy_s( outString, MAX_FILENAME_LEN, Filename );
					rv = 0;
					break;
				}
				else if ( !_wcsicmp( BaseName, L"ConEmuC.exe" ) ) {
					wcscpy_s( outString, MAX_FILENAME_LEN, Filename );
					rv = 0;
					break;
				}
			}
		}
	}

	free( ProcessListBuffer );
	return rv;
}



BOOL CALLBACK ConEmuWindowCheck( HWND hWnd, LPARAM lParam )
{
	if ( !lParam )
		return TRUE;


	wchar_t ClassName[SHORT_BUF_LEN] = L"";

	if ( !RealGetWindowClass( hWnd, ClassName, SHORT_BUF_LEN ) )
		return TRUE;

	if ( wcscmp( ClassName, ConEmuClass ) )
		return TRUE;

	HWND *ConEmuWindow = (HWND*) lParam;

	if ( *ConEmuWindow == NULL )
		*ConEmuWindow = hWnd;

	return TRUE;
}



HWND FindConEmuWindow()
{
	wchar_t Filename[MAX_FILENAME_LEN]			= L"";
	wchar_t CmdBuffer[MAX_FILENAME_LEN + 32]	= L"";
	HWND ConEmuHandle							= NULL;
	unsigned int Value							= 0;


	if ( GetConEmuCFilename( Filename ) )
		return NULL;

	wsprintf( CmdBuffer, L"\"%s\" /GUIMACRO IsConEmu > nul:", Filename );

	if ( Command( CmdBuffer, 0 ) )
		return NULL;


	Filename[0] = '\0';
	GetEnvironmentVariable( L"ConEmuHWND", Filename, MAX_FILENAME_LEN );
	if ( Filename[0] ) {
		if ( !ParseInt( Filename, &Value, NULL ) ) {
			Filename[0] = '\0';
			RealGetWindowClass( (HWND) Value, Filename, MAX_FILENAME_LEN );

			if ( !wcscmp( Filename, ConEmuClass ) )
				return (HWND) Value;
		}
	}

	EnumWindows( ConEmuWindowCheck, (LPARAM) &ConEmuHandle );

	return ConEmuHandle;
}



HWND MainWindow()
{
	HWND Test = NULL;

	if ( Console2WindowHandle )
		return Console2WindowHandle;
	
	Test = FindTakeCmdWindow();
	if ( Test )
		return Test;

	Test = FindConEmuWindow();
	if ( Test )
		return Test;
	
	return GetConsoleWindow();
}



int OpenSearchDialog( BOOL AutoSearch )
{
	if ( !SearchDialogHandle ) {
		SearchDialogHandle = CreateDialog( MyDLLHandle, MAKEINTRESOURCE( IDD_DIALOG1 ), GetConsoleWindow(), SearchDialogProc );

		if ( FirstOpen ) {
			if ( CleverPosition( SearchDialogHandle, MainWindow() ) )
				CenterWindowOnWindow( SearchDialogHandle, GetConsoleWindow() );

			FirstOpen = FALSE;
		}
		else if ( AlwaysBeClever ) {
			if ( CleverPosition( SearchDialogHandle, MainWindow() ) )
				SetWindowPos( SearchDialogHandle, HWND_TOP, DialogX, DialogY, 0, 0, SWP_NOSIZE | SWP_SHOWWINDOW );
		}
		else
			SetWindowPos( SearchDialogHandle, HWND_TOP, DialogX, DialogY, 0, 0, SWP_NOSIZE | SWP_SHOWWINDOW );

		if ( FadeOut ) {
			SetWindowLongPtr( SearchDialogHandle, GWL_EXSTYLE, GetWindowLongPtr( SearchDialogHandle, GWL_EXSTYLE ) | WS_EX_LAYERED );
			SetLayeredWindowAttributes( SearchDialogHandle, 0, FadeOut ? (BYTE) ActiveAlpha : ALPHA_OPAQUE, LWA_ALPHA );
		}
	}
	else
		PostMessage( SearchDialogHandle, WM_UPDATE, 0, 0 );

	ShowWindow( SearchDialogHandle, SW_SHOW );
	ForceForegroundWindow( SearchDialogHandle );

	if ( AutoSearch )
		TimerID = SetTimer( SearchDialogHandle, DEFAULT_TIMER_ID, AutoSearchTimeout, NULL );

	return 0;
}



int FixColor2()
{
	if ( ( ( HighlightColor1 & 0x00f0 ) >> 4 ) == ( HighlightColor1 & 0x000f ) )
		HighlightColor2 = DEFAULT_COLOR2;
	else {
		int bg = ( HighlightColor1 & 0x00f0 ) >> 4;
		int fg = 0x000f;

		bg = ( bg + 3 ) & 0x000f;
		if ( bg & 0x0008 )
			fg = 0x0000;

		HighlightColor2 = ( bg << 4 ) | fg;
	}

	return 0;
}



int RestoreOptions( HKEY MainKey )
{
	HKEY KeyHandle = NULL;
	DWORD keyval = 0xffff0000, errorval = 0xffff0000, cleverval = 0xffff0000;
	DWORD color1val = 0xffff0000, color2val = 0xffff0000;
	DWORD searchtypeval = 0xffff0000, wholewordsval = 0xffff0000, wrapallval = 0xffff0000;
	DWORD notcmdval		= 0x00000000;
	DWORD fadeoutval	= 0x00000000;
	DWORD activealphaval = 0xffff0000, inactivealphaval = 0xffff0000;

	DWORD buflen = 0;
	int rv = 1;


	if ( RegOpenKeyEx( MainKey, PluginRegKey, 0, KEY_READ, &KeyHandle ) != ERROR_SUCCESS )
		return 1;

	buflen = sizeof( DWORD );
	RegQueryValueEx( KeyHandle, HotkeyRegValue, NULL, NULL, (BYTE*) &keyval, &buflen );

	buflen = sizeof( DWORD );
	RegQueryValueEx( KeyHandle, NotFoundRegValue, NULL, NULL, (BYTE*) &errorval, &buflen );

	buflen = sizeof( DWORD );
	RegQueryValueEx( KeyHandle, CleverRegValue, NULL, NULL, (BYTE*) &cleverval, &buflen );

	buflen = sizeof( DWORD );
	RegQueryValueEx( KeyHandle, Color1RegValue, NULL, NULL, (BYTE*) &color1val, &buflen );

	buflen = sizeof( DWORD );
	RegQueryValueEx( KeyHandle, Color2RegValue, NULL, NULL, (BYTE*) &color2val, &buflen );

	buflen = sizeof( DWORD );
	RegQueryValueEx( KeyHandle, DefaultSearchTypeRegValue, NULL, NULL, (BYTE*) &searchtypeval, &buflen );

	buflen = sizeof( DWORD );
	RegQueryValueEx( KeyHandle, DefaultWholeWordsRegValue, NULL, NULL, (BYTE*) &wholewordsval, &buflen );

	buflen = sizeof( DWORD );
	RegQueryValueEx( KeyHandle, DefaultWrapAllRegValue, NULL, NULL, (BYTE*) &wrapallval, &buflen );

	buflen = sizeof( DWORD );
	RegQueryValueEx( KeyHandle, NoTakeCommandRegValue, NULL, NULL, (BYTE*) &notcmdval, &buflen );

	buflen = sizeof( DWORD );
	RegQueryValueEx( KeyHandle, FadeOutRegValue, NULL, NULL, (BYTE*) &fadeoutval, &buflen );

	buflen = sizeof( DWORD );
	RegQueryValueEx( KeyHandle, ActiveAlphaRegValue, NULL, NULL, (BYTE*) &activealphaval, &buflen );

	buflen = sizeof( DWORD );
	RegQueryValueEx( KeyHandle, InactiveAlphaRegValue, NULL, NULL, (BYTE*) &inactivealphaval, &buflen );

	if ( IsKnownKeyValue( keyval ) ) {
		HotKey = keyval;
			
		if ( ( errorval & 0xffff0000 ) == 0 )
			NotFoundMode = errorval;

		if ( ( cleverval & 0xffff0000 ) == 0 )
			AlwaysBeClever = cleverval;

		if ( ( color1val & 0xffff0000 ) == 0 )
			HighlightColor1 = color1val;

		if ( ( color2val & 0xffff0000 ) == 0 )
			HighlightColor2 = color2val;

		if ( ( searchtypeval == SEARCH_CASE_INSENSITIVE ) || ( searchtypeval == SEARCH_CASE_SENSITIVE ) || ( searchtypeval == SEARCH_REGEXP ) )
			SearchType = DefaultSearchType = searchtypeval;

		if ( ( wholewordsval & 0xffff0000 ) == 0 )
			WholeWordOnly = DefaultWholeWords = wholewordsval;

		if ( ( wrapallval & 0xffff0000 ) == 0 )
			WrapAllLines = DefaultWrapAll = wrapallval;

		NoLoadUnderTCmd = notcmdval;
		FadeOut = fadeoutval;

		if ( ( activealphaval >= ALPHA_MIN ) && ( activealphaval <= ALPHA_MAX ) && ( inactivealphaval >= ALPHA_MIN ) && ( inactivealphaval <= ALPHA_MAX ) ) {
			ActiveAlpha = activealphaval;
			InactiveAlpha = inactivealphaval;
		}

		FixColor2();
		rv = 0;
	}

	RegCloseKey( KeyHandle );
	return rv;
}



int RestoreOptions()
{
	if ( !RestoreOptions( HKEY_CURRENT_USER ) )
		return 0;

	if ( !RestoreOptions( HKEY_LOCAL_MACHINE ) )
		return 0;

	return 1;
}



BOOL CALLBACK Console2WindowCheck( HWND hWnd, LPARAM lParam )
{
	if ( !lParam )
		return TRUE;


	wchar_t ClassName[SHORT_BUF_LEN] = L"";

	if ( !RealGetWindowClass( hWnd, ClassName, SHORT_BUF_LEN ) )
		return TRUE;

	if ( wcscmp( ClassName, Console2Class ) )
		return TRUE;

	HWND *Console2hWnd = (HWND*) lParam;

	for ( int i = 0; i < MAX_CONSOLE2_WINDOWS; i++ ) {
		if ( Console2hWnd[i] == NULL ) {
			Console2hWnd[i] = hWnd;
			return TRUE;
		}
	}

	return TRUE;
}



int SameConsoleWindow( DWORD Pid )
{
	DWORD ConsoleWindowPid = 0;

	GetWindowThreadProcessId( GetConsoleWindow(), &ConsoleWindowPid );

	if ( ConsoleWindowPid == Pid )
		return 1;
	else
		return 0;
}



HWND FindConsole2Window()
{
	HWND Console2hWnd[MAX_CONSOLE2_WINDOWS];
	DWORD Console2Pid[MAX_CONSOLE2_WINDOWS];

	for ( int i = 0; i < MAX_CONSOLE2_WINDOWS; i++ )
		Console2hWnd[i] = NULL, Console2Pid[i] = 0;


	EnumWindows( Console2WindowCheck, (LPARAM) Console2hWnd );

	if ( Console2hWnd[0] == NULL )
		return NULL;

	for ( int i = 0; i < MAX_CONSOLE2_WINDOWS; i++ )
		if ( Console2hWnd[i] )
			GetWindowThreadProcessId( Console2hWnd[i], Console2Pid + i );


	DWORD MyPid = GetCurrentProcessId();
	DWORD Pid = MyPid, ParentPid = NULL, LastPid = NULL;

	do {
		ParentPid = GetParentPID( Pid );
		if ( ( ParentPid == Pid ) || ( ParentPid == 0xffffffff ) )
			break;

		LastPid = Pid;
		Pid = ParentPid;

		for ( int i = 0; i < MAX_CONSOLE2_WINDOWS; i++ ) {
			if ( Pid == Console2Pid[i] ) {
				if ( LastPid == MyPid )
					return Console2hWnd[i];
				else if ( SameConsoleWindow( LastPid ) )
					return Console2hWnd[i];
				else
					return NULL;
			}
		}

	} while ( TRUE );

	return NULL;
}



int RestorePluginDefaults()
{
	DefaultSearchType	= SEARCH_CASE_INSENSITIVE;
	DefaultWholeWords	= FALSE;
	DefaultWrapAll		= FALSE;

	HotKey				= DEFAULT_HOTKEY;
	NotFoundMode		= DEFAULT_NOT_FOUND;
	HighlightColor1		= DEFAULT_COLOR1;
	HighlightColor2		= DEFAULT_COLOR2;
	FirstOpen			= TRUE;
	AlwaysBeClever		= FALSE;
	FadeOut				= TRUE;
	ActiveAlpha			= DEFAULT_ACTIVE_ALPHA;
	InactiveAlpha		= DEFAULT_INACTIVE_ALPHA;

	if ( !SearchDialogHandle ) {
		SearchType			= SEARCH_CASE_INSENSITIVE;
		SearchText[0]		= '\0';
		WholeWordOnly		= FALSE;
		WrapAllLines		= FALSE;
		SearchCaret			= 0;
	}

	return 0;
}



BOOL APIENTRY DllMain( HANDLE hModule, DWORD  dwReason, LPVOID lpReserved )
{
	MyDLLHandle = (HINSTANCE) hModule;

	switch( dwReason )	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
	}
	return TRUE;
}



DLLExports LPPLUGININFO WINAPI GetPluginInfo( void )
{
	static PLUGININFO piInfo;

	piInfo.pszDll			= (LPTSTR) PLUGIN_NAME;
	piInfo.pszAuthor		= (LPTSTR) L"Charles Dye";
	piInfo.pszEmail			= (LPTSTR) L"cdye@unm.edu";
	piInfo.pszWWW			= (LPTSTR) L"http://charlesdye.net/plugins/consearch.html";
	piInfo.pszDescription	= (LPTSTR) L"Searches for text in the console buffer.";
	piInfo.pszFunctions		= (LPTSTR) PluginFeatures;
	piInfo.nMajor			= VER_MAJOR;
	piInfo.nMinor			= VER_MINOR;
	piInfo.nBuild			= VER_BUILD;

	return &piInfo;
}



DLLExports BOOL WINAPI InitializePlugin( void )
{
	InitHotkeys();

	wchar_t shortbuf[SHORT_BUF_LEN] = L"%[_transient]%[_pipe]%[noloadmsg]";
	ExpandVariables( shortbuf, 1 );

	RestorePluginDefaults();
	RestoreOptions();

	if ( NoLoadUnderTCmd )
		if ( FindTakeCmdWindow() ) {
			Printf( L"%s plugin: Take Command detected; not loading.\r\n", PLUGIN_NAME );
			return -1;
		}

	if ( !wcscmp( shortbuf, L"00" ) ) {
		GetKeyName( HotKey, shortbuf );
		Printf( L"%s v%d.%d.%d loaded; hotkey is %s.\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, shortbuf );
	}


	MyThread = GetCurrentThreadId();
	//  wprintf( L"   * MyDLLHandle = 0x%08x   MyThread = 0x%08x\n", MyDLLHandle, MyThread );

	HookHandle = SetWindowsHookEx( WH_GETMESSAGE, MyMessageFilter, MyDLLHandle, MyThread );
	//  wprintf( L"   * Installed message filter hook 0x%08x\n", HookHandle );

	Console2WindowHandle = FindConsole2Window();
	//  wprintf( L"   * Console2WindowHandle = 0x%08x\n", Console2WindowHandle );

	FirstOpen = TRUE;
	return 0;
}



DLLExports BOOL WINAPI ShutdownPlugin( BOOL bEndProcess )
{
	if ( SearchDialogHandle )
		DestroyWindow( SearchDialogHandle );

	if ( HookHandle ) {
		UnhookWindowsHookEx( HookHandle );
		//  wprintf( L"   * Removed message filter hook 0x%08x\n", HookHandle );
	}

	return 0;
}



DLLExports int WINAPI key( LPKEYINFO24 ki )
{
	if ( !HotKey )
		return 0;


	if ( ki->nKey == 0 ) {
		if ( !NewKeyHook )
			return 0;
		else if ( ki->pszKey == NULL )
			return 0;
	}


	int inKey = ki->nKey;
	int shiftKeys = 0;

	if ( NewKeyHook ) {
		if ( !inKey && ki->pszKey ) {
			inKey = ParseKeyName24( ki->pszKey );

			if ( !inKey )
				return 0;
		}
	}


	if ( GetAsyncKeyState( VK_SHIFT ) & 0x8000 )
		shiftKeys |= 1;

	if ( GetAsyncKeyState( VK_CONTROL ) & 0x8000 )
		shiftKeys |= 2;

	if ( GetAsyncKeyState( VK_MENU ) & 0x8000 )
		shiftKeys |= 4;

	if ( shiftKeys == 3 )
		inKey |= 0x20000;
	else if ( shiftKeys == 6 )
		inKey |= 0x40000;

	if ( shiftKeys == 3 )
		inKey |= 0x20000;
	else if ( shiftKeys == 6 )
		inKey |= 0x40000;
	else if ( NewKeyHook && ( shiftKeys == 7 ) )
		inKey |= 0x60000;


	if ( inKey != HotKey )
		return 0;

	ki->nKey = 0;
	if ( NewKeyHook )
		ki->pszKey[0] = '\0';

	OpenSearchDialog( FALSE );
	return 0;
}



#define PARSE_BREAK_SPACES		1
#define PARSE_BREAK_COMMAS		2
#define PARSE_SLASHES_KLUDGE	4
#define	PARSE_QUOTES_KLUDGE		8
#define PARSE_EQUALS_KLUDGE		16
#define PARSE_ONE_ARG_KLUDGE	32
#define PARSE_RETAIN_QUOTES		64
#define PARSE_FORCE_UPPERCASE	128
#define PARSE_NO_TRIM			256
#define PARSE_IGNORE_QUOTES		512
#define PARSE_EQ_QT_KLUDGE		1024


int ParseArgs( LPTSTR inString, unsigned int flags )
{
	/*
		flags:  bit 0 - break args at spaces
				bit 1 - break args at commas (either bit 0 or bit 1 must be set)
				bit 2 - slashes kludge:   treat /a/b like /a /b
				bit 3 - quotes kludge:    treat /a"foo" like /a:"foo"
				bit 4 - equals kludge:    break at first unquoted equals sign
				bit 5 - one-arg kludge:   if arg does not begin with /, allow unquoted spaces
				bit 6 - don't swallow double quotes
				bit 7 - force uppercase
				bit 8 - don't trim spaces from the right
				bit 9 - disable all special handling of double quotes
	*/

	if ( !inString )
		return -1;

	numargs = 0;

	int argnum = 0;
	int i = 0;
	int a = 0;
	int b = 0;
	wchar_t ch = '\0';
	BOOL quote  = FALSE;
	BOOL adv    = FALSE;
	BOOL done   = FALSE;
	int eqkl    = 0;

	if ( ( flags & ( PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS | PARSE_EQUALS_KLUDGE ) ) == 0 )
		flags |= PARSE_BREAK_SPACES;

	for ( i = 0; i < MAX_ARGS; i++ ) {
		arg[i] = NULL;
		arglen[i] = 0;
	}

	for ( i = 0; !done; i++ ) {
		ch = inString[i];
		adv = FALSE;

		if ( ( ch == '"' ) && ( ( flags & PARSE_IGNORE_QUOTES ) == 0 ) ) {
			quote = !quote;

			if ( quote && ( a == 2 ) && ( arg[argnum][0] == '/' ) && ( iswalpha( arg[argnum][1] ) ) && ( flags & PARSE_QUOTES_KLUDGE ) ) {
				if ( flags & PARSE_RETAIN_QUOTES ) {
					if ( b < ( MAX_ARG_LEN - 1 ) )
						allargs[b++] = ':';
				}
				else
					ch = ':';
			}
			else if ( ( flags & PARSE_RETAIN_QUOTES ) == 0 )
				continue;
		}

		if ( iswspace( ch ) && ( a == 0 ) && !quote )
			continue;

		if ( ( a == 0 ) && ( ch != '/' ) && ( ( flags & ( PARSE_BREAK_SPACES | PARSE_ONE_ARG_KLUDGE ) ) == ( PARSE_BREAK_SPACES | PARSE_ONE_ARG_KLUDGE ) ) )
			flags &= ~PARSE_BREAK_SPACES;

		if ( iswspace( ch ) && !quote && ( a == 1 ) && ( argnum > 0 ) && ( flags & PARSE_EQUALS_KLUDGE ) && ( arg[argnum][0] == '=' ) )
			continue;

		if ( iswspace( ch ) && !quote && ( flags & PARSE_BREAK_SPACES ) )
			adv = TRUE;

		if ( ( ch == ',' ) && !quote && ( flags & PARSE_BREAK_COMMAS ) )
			adv = TRUE;

		if ( ( ch == '/' ) && !quote && ( a > 1 ) && ( arg[argnum][0] == '/' ) && ( flags & PARSE_SLASHES_KLUDGE ) ) {
			argnum++;
			a = 0;
			b++;
			allargs[b] = '\0';
		}

		if ( ( ch == '=' ) && !quote && ( eqkl == 0 ) && ( a > 0 ) && ( arg[argnum][0] != '/' ) && ( flags & PARSE_EQUALS_KLUDGE ) ) {
			argnum++;
			a = 0;
			b++;
			allargs[b] = '\0';
			eqkl = argnum;
			flags &= ~( PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS );
			//- Printf( L"   * equals kludge invoked; eqkl = %d\n", eqkl );

			if ( flags & PARSE_EQ_QT_KLUDGE )
				flags |= PARSE_IGNORE_QUOTES;
		}
		else if ( ( ch == '=' ) && !quote && ( flags & PARSE_EQUALS_KLUDGE ) ) {
			eqkl = -1;
			flags &= ~( PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS );

			if ( flags & PARSE_EQ_QT_KLUDGE )
				flags |= PARSE_IGNORE_QUOTES;
		}

		if ( ch == '\0' ) {
			if ( a )
				argnum++;

			done = TRUE;
			continue;
		}

		if ( adv ) {
			argnum++;
			a = 0;
			b++;
			allargs[b] = '\0';
			quote = FALSE;
			continue;
		}

		if ( b >= ( MAX_ARG_LEN - 1 ) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Buffer overflow: b = %d, MAX_ARG_LEN = %d \r\n", b, MAX_ARG_LEN );
			return -1;
		}

		if ( argnum < MAX_ARGS ) {
			if ( a == 0 )
				arg[argnum] = allargs + b;

			if ( flags & PARSE_FORCE_UPPERCASE )
				ch = towupper( ch );

			a++;
			arglen[argnum]  = a;
			allargs[b++] = ch;
			allargs[b]   = '\0';
		}

	}

	if ( ( flags & PARSE_NO_TRIM ) == 0 ) {
		for ( i = 0; i < argnum; i++ ) {
			while ( ( arglen[i] > 0 ) && iswspace( arg[i][arglen[i] - 1] ) ) {
				arglen[i]--;
				arg[i][arglen[i]] = '\0';
			}
		}
	}

	numargs = argnum;
	return argnum;
}



int IsColonOrEquals( wchar_t ch )
{
	if ( ( ch == ':' ) || ( ch == '=' ) )
		return 1;

	return 0;
}



int SaveOptions()
{
	int rv = 0;
	HKEY KeyHandle = NULL;

	if ( RegCreateKeyEx( HKEY_CURRENT_USER, PluginRegKey, 0, NULL, 0, KEY_READ | KEY_WRITE, NULL, &KeyHandle, NULL ) == ERROR_SUCCESS ) {

		if ( RegSetValueEx( KeyHandle, HotkeyRegValue, 0, REG_DWORD, (BYTE*) &HotKey, sizeof( DWORD ) ) != ERROR_SUCCESS )
			rv = 2;

		if ( RegSetValueEx( KeyHandle, NotFoundRegValue, 0, REG_DWORD, (BYTE*) &NotFoundMode, sizeof( DWORD ) ) != ERROR_SUCCESS )
			rv = 3;

		if ( RegSetValueEx( KeyHandle, CleverRegValue, 0, REG_DWORD, (BYTE*) &AlwaysBeClever, sizeof( DWORD ) ) != ERROR_SUCCESS )
			rv = 4;

		if ( RegSetValueEx( KeyHandle, Color1RegValue, 0, REG_DWORD, (BYTE*) &HighlightColor1, sizeof( DWORD ) ) != ERROR_SUCCESS )
			rv = 5;

		if ( RegSetValueEx( KeyHandle, Color2RegValue, 0, REG_DWORD, (BYTE*) &HighlightColor2, sizeof( DWORD ) ) != ERROR_SUCCESS )
			rv = 6;

		if ( RegSetValueEx( KeyHandle, DefaultSearchTypeRegValue, 0, REG_DWORD, (BYTE*) &DefaultSearchType, sizeof( DWORD ) ) != ERROR_SUCCESS )
			rv = 7;

		if ( RegSetValueEx( KeyHandle, DefaultWholeWordsRegValue, 0, REG_DWORD, (BYTE*) &DefaultWholeWords, sizeof( DWORD ) ) != ERROR_SUCCESS )
			rv = 8;

		if ( RegSetValueEx( KeyHandle, DefaultWrapAllRegValue, 0, REG_DWORD, (BYTE*) &DefaultWrapAll, sizeof( DWORD ) ) != ERROR_SUCCESS )
			rv = 9;

		if ( RegSetValueEx( KeyHandle, NoTakeCommandRegValue, 0, REG_DWORD, (BYTE*) &NoLoadUnderTCmd, sizeof( DWORD ) ) != ERROR_SUCCESS )
			rv = 10;

		if ( RegSetValueEx( KeyHandle, FadeOutRegValue, 0, REG_DWORD, (BYTE*) &FadeOut, sizeof( DWORD ) ) != ERROR_SUCCESS )
			rv = 11;

		if ( RegSetValueEx( KeyHandle, ActiveAlphaRegValue, 0, REG_DWORD, (BYTE*) &ActiveAlpha, sizeof( DWORD ) ) != ERROR_SUCCESS )
			rv = 12;

		if ( RegSetValueEx( KeyHandle, InactiveAlphaRegValue, 0, REG_DWORD, (BYTE*) &InactiveAlpha, sizeof( DWORD ) ) != ERROR_SUCCESS )
			rv = 13;

		RegCloseKey( KeyHandle );
	}
	else
		rv = 1;

	return rv;
}



int GetHighlightColorName( unsigned int Color, LPTSTR outString )
{
	if ( !outString )
		return -1;


	if ( ( ( Color & 0x00f0 ) >> 4 ) == ( Color & 0x000f ) ) {
		wcscpy_s( outString, SHORT_BUF_LEN, L"Reverse video" );
		return 0;
	}

	wsprintf( outString, L"%s%s on %s%s", ( Color & 0x0008 ) ? L"bright " : L"", ColorNames[ Color & 0x0007 ], ( Color & 0x0080 ) ? L"bright " : L"", ColorNames[ ( Color & 0x0070 ) >> 4 ] );
	outString[0] = towupper( outString[0] );
	return 0;
}



int GetSearchType( unsigned int SearchType, LPTSTR outString )
{
	if ( !outString )
		return -1;

	switch ( SearchType ) {
		case SEARCH_CASE_INSENSITIVE :
			wcscpy_s( outString, SHORT_BUF_LEN, L"Case insensitive" );
			break;

		case SEARCH_CASE_SENSITIVE :
			wcscpy_s( outString, SHORT_BUF_LEN, L"Case sensitive" );
			break;

		case SEARCH_REGEXP :
			wcscpy_s( outString, SHORT_BUF_LEN, L"Regular expression" );
			break;

		default :
			wcscpy_s( outString, SHORT_BUF_LEN, L"Unknown" );
	}

	return 0;
}



int GetNotFoundMode( unsigned int NotFoundMode, LPTSTR outString )
{
	if ( !outString )
		return -1;

	if ( NotFoundMode & NOT_FOUND_MESSAGEBOX )
		wcscpy_s( outString, SHORT_BUF_LEN, L"Message box" );
	else if ( NotFoundMode & NOT_FOUND_CLOSE )
		wcscpy_s( outString, SHORT_BUF_LEN, L"Close dialog" );
	else {
		switch ( NotFoundMode & NOT_FOUND_SOUNDS ) {
			case NOT_FOUND_ASTERISK :
				wcscpy_s( outString, SHORT_BUF_LEN, L"Asterisk sound" );
				break;

			case NOT_FOUND_DEFAULT_BEEP :
				wcscpy_s( outString, SHORT_BUF_LEN, L"Default beep sound" );
				break;

			case NOT_FOUND_EXCLAMATION :
				wcscpy_s( outString, SHORT_BUF_LEN, L"Exclamation sound" );
				break;

			default :
				wcscpy_s( outString, SHORT_BUF_LEN, L"Hand sound" );
				break;
		}
	}

	return 0;
}



DLLExports int WINAPI consearchopts( LPTSTR inString )
{
	if ( !inString )
		return -1;


	BOOL Save = FALSE, Restore = FALSE, PluginDefaults = FALSE, List = FALSE;
	int NewHotkey			= -1;
	int NewNotFoundMode		= -1;
	int NewClever			= -1;
	int NewNoLoadMode		= -1;
	int NewFadeOut			= -1;
	int NewActiveAlpha		= -1;
	int NewInactiveAlpha	= -1;

	
	wchar_t shortbuf[SHORT_BUF_LEN] = L"";
	int rv = 0;


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_FORCE_UPPERCASE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcscmp( arg[i], L"/?" ) )
				return ShowCmdHelp( ConSearchOptsHelpText, FALSE );
			else if ( !wcscmp( arg[i], L"/??" ) )
				return ShowCmdHelp( ConSearchOptsHelpText, TRUE );
			else if ( !wcsncmp( arg[i], L"/A", 2 ) ) {
				wchar_t *ch		= arg[i] + 2;
				if ( IsColonOrEquals( *ch ) )
					ch++;

				if ( *ch == '\0' ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Missing ActiveAlpha value: \"%s\"\r\n", arg[i] );
					return 1;
				}
				else if ( ParseInt( ch, (unsigned int*) &NewActiveAlpha, &ch ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad ActiveAlpha value: \"%s\"\r\n", ch );
					return 1;
				}

				if ( ( NewActiveAlpha < ALPHA_MIN ) || ( NewActiveAlpha > ALPHA_MAX ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"ActiveAlpha must be %u - %u: %u\r\n", ALPHA_MIN, ALPHA_MAX, NewActiveAlpha );
					return 1;
				}

				if ( *ch == '\0' )
					NewInactiveAlpha = NewActiveAlpha;
				else if ( ( *ch == ',' ) || ( *ch == ';' ) ) {

					ch++;

					if ( *ch == '\0' ) {
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Missing InactiveAlpha value: \"%s\"\r\n", arg[i] );
						return 1;
					}
					else if ( ParseInt( ch, (unsigned int*) &NewInactiveAlpha, NULL ) ) {
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Bad InactiveAlpha value: \"%s\"\r\n", ch );
						return 1;
					}

					if ( ( NewInactiveAlpha < ALPHA_MIN ) || ( NewInactiveAlpha > ALPHA_MAX ) ) {
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"InactiveAlpha must be %u - %u: %u\r\n", ALPHA_MIN, ALPHA_MAX, NewInactiveAlpha );
						return 1;
					}
				}
				else {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", ch );
					return 1;
				}
			}
			else if ( !wcscmp( arg[i], L"/D" ) )
				PluginDefaults = TRUE;
			else if ( !wcsncmp( arg[i], L"/DA", 3 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] == 'Y' )
					DefaultWrapAll = TRUE;
				else if ( arg[i][j] == 'N' )
					DefaultWrapAll = FALSE;
				else {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Should be Y or N: \"%s\"\r\n", arg[i] + j );
					return 1;
				}
			}
			else if ( !wcsncmp( arg[i], L"/DW", 3 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] == 'Y' )
					DefaultWholeWords = TRUE;
				else if ( arg[i][j] == 'N' )
					DefaultWholeWords = FALSE;
				else {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Should be Y or N: \"%s\"\r\n", arg[i] + j );
					return 1;
				}
			}
			else if ( !wcsncmp( arg[i], L"/E", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				switch ( arg[i][j] ) {
					case 'A' :
						NewNotFoundMode = NOT_FOUND_ASTERISK;
						break;

					case 'C' :
						NewNotFoundMode = NOT_FOUND_CLOSE;
						break;

					case 'D' :
						NewNotFoundMode = NOT_FOUND_DEFAULT_BEEP;
						break;

					case 'E' :
						NewNotFoundMode = NOT_FOUND_EXCLAMATION;
						break;

					case 'H' :
						NewNotFoundMode = NOT_FOUND_HAND;
						break;

					case 'M' :
						NewNotFoundMode = NOT_FOUND_MESSAGEBOX;
						break;

					default :
						DisplayErrorHeader();
						Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Unknown error mode: \"%s\"\r\n", arg[i] );
						return 1;
				}
			}
			else if ( !wcsncmp( arg[i], L"/F", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] == 'Y' )
					NewFadeOut = TRUE;
				else if ( arg[i][j] == 'N' )
					NewFadeOut = FALSE;
				else {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Should be Y or N: \"%s\"\r\n", arg[i] + j );
					return 1;
				}
			}
			else if ( !wcsncmp( arg[i], L"/H", 2 ) ) {
				wchar_t *ch = arg[i] + 2;
				if ( IsColonOrEquals( *ch ) )
					ch++;
				
				unsigned int test = 0;
				int fg = -1, bg = -1;

				if ( !_wcsnicmp( ch, L"inv", 3 ) || !_wcsnicmp( ch, L"rev", 3 ) ) {
					HighlightColor1 = 0x0000;
					HighlightColor2 = DEFAULT_COLOR2;
				}
				else if ( !ParseInt( ch, &test, NULL ) ) {
					if ( ( test > 0xff ) || ( test && ( ( ( test & 0x00f0 ) >> 4 ) == ( test & 0x000f ) ) ) ) {
						DisplayErrorHeader();
						Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad highlight color: \"%s\"\r\n", arg[i] );
						return 1;
					}

					HighlightColor1 = test;
					FixColor2();
				}
				else {
					ParseColors( ch, &fg, &bg );

					if ( ( fg < 0 ) || ( bg < 0 ) || ( fg == bg ) ) {
						DisplayErrorHeader();
						Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad highlight colors: \"%s\"\r\n", arg[i] );
						return 1;
					}

					HighlightColor1 = ( bg << 4 ) | fg;
					FixColor2();
				}
			}
			else if ( !wcsncmp( arg[i], L"/K", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( !wcscmp( arg[i] + j, L"NONE" ) )
					NewHotkey = 0;
				else {
					NewHotkey = ParseKeyName( arg[i] + j );

					if ( NewHotkey == 0 ) {
						DisplayErrorHeader();
						Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Unknown key name \"%s\"\r\n", arg[i] );
						return 1;
					}
					else if ( NewHotkey == -1 ) {
						DisplayErrorHeader();
						Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Invalid hotkey \"%s\"\r\n", arg[i] );
						return 1;
					}
					else if ( NewHotkey == -2 ) {
						DisplayErrorHeader();
						Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Key not allowed \"%s\"\r\n", arg[i] );
						return 1;
					}
				}
			}
			else if ( !wcscmp( arg[i], L"/L" ) )
				List = TRUE;
			else if ( !wcscmp( arg[i], L"/R" ) )
				Restore = TRUE;
			else if ( !wcscmp( arg[i], L"/S" ) )
				Save = TRUE;
			else if ( !wcsncmp( arg[i], L"/T", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				switch ( arg[i][j] ) {
					case 'I' :
						DefaultSearchType = SEARCH_CASE_INSENSITIVE;
						break;

					case 'R' :
						DefaultSearchType = SEARCH_REGEXP;
						break;

					case 'S' :
						DefaultSearchType = SEARCH_CASE_SENSITIVE;
						break;

					default :
						DisplayErrorHeader();
						Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Unknown search type: \"%s\"\r\n", arg[i] );
						return 1;
				}
			}
			else if ( !wcsncmp( arg[i], L"/W", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] == 'Y' )
					NewClever = 1;
				else if ( arg[i][j] == 'N' )
					NewClever = 0;
				else {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Should be Y or N: \"%s\"\r\n", arg[i] + j );
					return 1;
				}
			}
			else if ( !wcsncmp( arg[i], L"/X", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] == 'Y' )
					NewNoLoadMode = 0;
				else if ( arg[i][j] == 'N' )
					NewNoLoadMode = 1;
				else {
					DisplayErrorHeader();
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Should be Y or N: \"%s\"\r\n", arg[i] + j );
					return 1;
				}
			}
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Unknown option \"%s\"\r\n", arg[i] );
				return 1;
			}
			else {
				DisplayErrorHeader();
				Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Syntax error \"%s\"\r\n", arg[i] );
				return 1;
			}
		}
	}


	if ( List ) {
		if ( Save || Restore || PluginDefaults ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Can\u2019t combine /L with other options\r\n" );
			return 1;
		}

		ListLegalKeys();
		return 0;
	}


	if ( PluginDefaults && Restore ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Can\u2019t combine /D and /R\r\n" );
		return 1;
	}


	if ( Restore )
		RestoreOptions();
	else if ( PluginDefaults )
		RestorePluginDefaults();


	if ( NewHotkey != -1 )
		HotKey = NewHotkey;

	if ( NewNotFoundMode != -1 )
		NotFoundMode = NewNotFoundMode;

	if ( NewClever != -1 )
		AlwaysBeClever = NewClever;

	if ( NewNoLoadMode != -1 )
		NoLoadUnderTCmd = NewNoLoadMode;

	if ( NewFadeOut != -1 )
		FadeOut = NewFadeOut;

	if ( NewActiveAlpha != -1 )
		ActiveAlpha = NewActiveAlpha;

	if ( NewInactiveAlpha != -1 )
		InactiveAlpha = NewInactiveAlpha;


	if ( SearchDialogHandle && ( Restore || PluginDefaults || ( NewFadeOut != -1 ) || ( NewActiveAlpha != -1 ) || ( NewInactiveAlpha != -1 ) ) ) {
		if ( FadeOut ) {
			SetWindowLongPtr( SearchDialogHandle, GWL_EXSTYLE, GetWindowLongPtr( SearchDialogHandle, GWL_EXSTYLE ) | WS_EX_LAYERED );
			SetLayeredWindowAttributes( SearchDialogHandle, 0, DialogActive ? (BYTE) ActiveAlpha : (BYTE) InactiveAlpha, LWA_ALPHA );
		}
		else {
			SetLayeredWindowAttributes( SearchDialogHandle, 0, ALPHA_OPAQUE, 0 );
			SetWindowLongPtr( SearchDialogHandle, GWL_EXSTYLE, GetWindowLongPtr( SearchDialogHandle, GWL_EXSTYLE ) & ~WS_EX_LAYERED );
		}
	}


	Printf( L"\r\nConSearch v%u.%u.%u options:\r\n\r\n", VER_MAJOR, VER_MINOR, VER_BUILD );

	GetKeyName( HotKey, shortbuf );
	Printf( L"ConSearch hotkey ........... /K : %s\r\n", shortbuf );

	GetHighlightColorName( HighlightColor1, shortbuf );
	Printf( L"Highlight color ............ /H : %s\r\n", shortbuf );

	GetNotFoundMode( NotFoundMode, shortbuf );
	Printf( L"Error indicator ............ /E : %s\r\n", shortbuf );
	
	Printf( L"Dialog near console ........ /W : %s\r\n", AlwaysBeClever ? L"Yes" : L"No" );

	GetSearchType( DefaultSearchType, shortbuf );
	Printf( L"Default search type ........ /T : %s\r\n", shortbuf );

	Printf( L"Fade when inactive ......... /F : %s\r\n", FadeOut ? L"Yes" : L"No" );
	Printf( L"Load in Take Command tab ... /X : %s\r\n", NoLoadUnderTCmd ? L"No" : L"Yes" );

	if ( Save ) {
		if ( SaveOptions() )
			rv = 2;
		else
			QPuts( L"\r\nSettings saved to the registry.\r\n" );
	}

	return rv;
}



DLLExports int WINAPI consearch( LPTSTR inString )
{
	if ( !inString )
		return -1;

	LPTSTR NewSearchText	= NULL;
	BOOL AutoSearch			= FALSE;
	BOOL CloseDialog		= FALSE;
	BOOL OtherOption		= FALSE;


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_ONE_ARG_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcscmp( arg[i], L"/?" ) )
				return ShowCmdHelp( ConSearchHelpText, FALSE );
			else if ( !wcscmp( arg[i], L"/??" ) )
				return ShowCmdHelp( ConSearchHelpText, TRUE );
			else if ( !_wcsnicmp( arg[i], L"/A", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( towupper( arg[i][j] ) == 'Y' )
					WrapAllLines = TRUE;
				else if ( towupper( arg[i][j] ) == 'N' )
					WrapAllLines = FALSE;
				else {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Should be Y or N: \"%s\"\r\n", arg[i] + j );
					return 1;
				}

				OtherOption = TRUE;
			}
			else if ( !_wcsicmp( arg[i], L"/C" ) )
				CloseDialog = TRUE;
			else if ( !_wcsicmp( arg[i], L"/I" ) )
				SearchType = SEARCH_CASE_INSENSITIVE, OtherOption = TRUE;
			else if ( !_wcsicmp( arg[i], L"/R" ) )
				SearchType = SEARCH_REGEXP, OtherOption = TRUE;
			else if ( !_wcsicmp( arg[i], L"/S" ) )
				SearchType = SEARCH_CASE_SENSITIVE, OtherOption = TRUE;
			else if ( !_wcsnicmp( arg[i], L"/W", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( towupper( arg[i][j] ) == 'Y' )
					WholeWordOnly = 1;
				else if ( towupper( arg[i][j] ) == 'N' )
					WholeWordOnly = 0;
				else {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Should be Y or N: \"%s\"\r\n", arg[i] + j );
					return 1;
				}

				OtherOption = TRUE;
			}
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Unknown option \"%s\"\r\n", arg[i] );
				return 1;
			}
			else if ( !NewSearchText )
				NewSearchText = arg[i];
			else {
				DisplayErrorHeader();
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Syntax error \"%s\"\r\n", arg[i] );
				return 1;
			}
		}
	}


	if ( CloseDialog ) {
		if ( OtherOption || NewSearchText ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Can\u2019t combine /C with other options\r\n" );
			return 2;
		}

		if ( !SearchDialogHandle ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Search dialog is not open\r\n" );
			return 2;
		}

		DestroyWindow( SearchDialogHandle );
		return 0;
	}


	SearchText[0] = '\0';

	if ( NewSearchText ) {
		wcscpy_s( SearchText, MAX_ARG_LEN, NewSearchText );
		AutoSearch = TRUE;
	}

	SearchCaret = 0;
	OpenSearchDialog( AutoSearch );
	return 0;
}



LPTSTR UsageText( const wchar_t *inString )
{
	if ( !inString )
		return NULL;


	wchar_t *ch = wcschr( (wchar_t*) inString, '\n' );


	if ( !ch )
		return NULL;

	while ( *ch == '\n' )
		ch++;

	return ch;
}



DLLExports int WINAPI Help( LPTSTR inString )
{
	if ( !_wcsicmp( inString, L"consearch" ) ) {
		MessageBox( GetForegroundWindow(), ConSearchHelpText, HelpWindowTitle, MB_OK | MB_SETFOREGROUND );
		return 1;
	}
	else if ( !_wcsicmp( inString, L"consearchopts" ) ) {
		MessageBox( GetForegroundWindow(), ConSearchOptsHelpText, HelpWindowTitle, MB_OK | MB_SETFOREGROUND );
		return 1;
	}
	else
		return 0;
}



DLLExports int WINAPI Usage( LPTSTR inString, LPTSTR outString )
{
	if ( !_wcsicmp( inString, L"consearch" ) ) {
		wcscpy_s( outString, SHORT_BUF_LEN, UsageText( ConSearchHelpText ) );
		return 1;
	}
	else if ( !_wcsicmp( inString, L"consearchopts" ) ) {
		wcscpy_s( outString, SHORT_BUF_LEN, UsageText( ConSearchOptsHelpText ) );
		return 1;
	}
	else
		return 0;
}

